@@ -1,9 +1,13 @@
 package DBIx::Array;
 use strict;
 use warnings;
+use base qw{Package::New};
+use Data::Dumper qw{Dumper};
+use List::Util qw(sum);
 use DBI;
+use DBIx::Array::Session::Action;
 
-our $VERSION='0.26';
+our $VERSION='0.47';
 our $PACKAGE=__PACKAGE__;
 
 =head1 NAME
@@ -29,25 +33,28 @@ With stored connection information from a File
 
 =head1 DESCRIPTION
 
-This module is for people who truly understand SQL and who understand Perl data structures.  If you understand how to modify your SQL to meet your data requirements then this module is for you.  In the example below, only one line of code is needed to generate an entire HTML table.
+This module provides a Perl data structure interface for Structured Query Language (SQL).  This module is for people who truly understand SQL and who understand Perl data structures.  If you understand how to modify your SQL to meet your data requirements then this module is for you.
 
-  print &tablename($dba->sqlarrayarrayname(&sql, 15)), "\n";
-
-  sub tablename {
-    use CGI; my $html=CGI->new(""); #you would pass this reference
-    return $html->table($html->Tr([map {$html->td($_)} @_]));
-  }
-   
-  sub sql { #Oracle SQL
-    return q{SELECT LEVEL AS "Number",
-                    TRIM(TO_CHAR(LEVEL, 'rn')) as "Roman Numeral"
-               FROM DUAL CONNECT BY LEVEL <= ? ORDER BY LEVEL};
-  }
-
-This module is used to connect to both Oracle 10g and 11g using L<DBD::Oracle> on both Linux and Win32, MySQL 4 and 5 using L<DBD::mysql> on Linux, and Microsoft SQL Server using L<DBD::Sybase> on Linux and using L<DBD::ODBC> on Win32 systems in a 24x7 production environment.  The tests are written against L<DBD::CSV> and L<DBD::XBase>.
+This module is used to connect to Oracle 10g and 11g using L<DBD::Oracle> on both Linux and Win32, MySQL 4 and 5 using L<DBD::mysql> on Linux, Microsoft SQL Server using L<DBD::Sybase> on Linux and using L<DBD::ODBC> on Win32 systems, and PostgreSQL using L<DBD::Pg> in a 24x7 production environment.  Tests are written against L<DBD::CSV> and L<DBD::XBase>.
 
 =head1 USAGE
 
+Loop through data
+
+  foreach my $row ($dbx->sqlarrayhash($sql, @bind)) {
+    do_something($row->{"id"}, $row->{"column"});
+  }
+
+Easily generate an HTML table
+
+  my $cgi  = CGI->new("");
+  my $html = $cgi->table($cgi->Tr([map {$cgi->td($_)} $dbx->sqlarrayarrayname($sql, @param)]));
+
+Bless directly into a class
+
+  my ($object) = $dbx->sqlarrayobject("My::Package", $sql, {id=>$id}); #bless({id=>1, name=>'foo'}, 'My::Package');
+  my @objects  = $dbx->absarrayobject("My::Package", "myview", '*', {active=>1}, ["name"]); #($object, $object, ...)
+
 =head1 CONSTRUCTOR
 
 =head2 new
@@ -59,34 +66,34 @@ This module is used to connect to both Oracle 10g and 11g using L<DBD::Oracle> o
 
 =cut
 
-sub new {
-  my $this=shift;
-  my $class=ref($this) || $this;
-  my $self={};
-  bless $self, $class;
-  $self->initialize(@_);
-  return $self;
-}
+#See Package::New->new
 
-=head1 METHODS
+=head1 METHODS (Properties)
 
-=head2 initialize
+=head2 dbh
+
+Sets or returns the database handle object.
+
+  my $dbh=$dbx->dbh;
+  $dbx->dbh($dbh);  #if you already have a connection
 
 =cut
 
-sub initialize {
+sub dbh {
   my $self=shift;
-  %$self=@_;
+  if (@_) {
+    $self->{'dbh'}=shift;
+    CORE::delete $self->{'_prepared'}; #clear cache if we switch handles
+  }
+  return $self->{'dbh'};
 }
 
-=head1 METHODS (Properties)
-
 =head2 name
 
 Sets or returns a user friendly identification string for this database connection
 
   my $name=$dbx->name;
-  my $name=$dbx->name($string);
+  $dbx->name($string);
 
 =cut
 
@@ -100,18 +107,15 @@ sub name {
 
 =head2 connect
 
-Connects to the database and returns the database handle.
+Wrapper around DBI->connect; Connects to the database, sets dbh property, and returns the database handle.
 
-  $dbx->connect($connection, $user, $pass, \%opt);
+  $dbx->connect($connection, $user, $pass, \%opt); #sets $dbx->dbh
+  my $dbh=$dbx->connect($connection, $user, $pass, \%opt);
 
-Pass through to DBI->connect;
-
-Examples: 
+Examples:
 
   $dbx->connect("DBI:mysql:database=mydb;host=myhost", "user", "pass", {AutoCommit=>1, RaiseError=>1});
-
   $dbx->connect("DBI:Sybase:server=myhost;datasbase=mydb", "user", "pass", {AutoCommit=>1, RaiseError=>1}); #Microsoft SQL Server API is same as Sybase API
-
   $dbx->connect("DBI:Oracle:TNSNAME", "user", "pass", {AutoCommit=>1, RaiseError=>1});
 
 =cut
@@ -119,27 +123,30 @@ Examples:
 sub connect {
   my $self=shift;
   my $dbh=DBI->connect(@_);
-  return $self->dbh($dbh);
+  $self->dbh($dbh);
+  CORE::delete $self->{'action'} if exists $self->{'action'};
+  tie $self->{'action'}, "DBIx::Array::Session::Action", (parent=>$self);
+  return $self->dbh;
 }
 
 =head2 disconnect
 
-Calls $dbh->disconnect
+Wrapper around dbh->disconnect
 
   $dbx->disconnect;
 
-Pass through to dbh->disconnect
-
 =cut
 
 sub disconnect {
   my $self=shift;
+  untie $self->{'action'};
+  CORE::delete $self->{'action'};
   return $self->dbh->disconnect
 }
 
 =head2 commit
 
-Pass through to dbh->commit
+Wrapper around dbh->commit
 
   $dbx->commit;
 
@@ -152,7 +159,7 @@ sub commit {
 
 =head2 rollback
 
-Pass through to dbh->rollback
+Wrapper around dbh->rollback
 
   $dbx->rollback;
 
@@ -163,9 +170,25 @@ sub rollback {
   return $self->dbh->rollback;
 }
 
+=head2 prepare
+
+Wrapper around dbh->prepare with a local cache.
+
+  my $sth=$dbh->prepare($sql);
+
+=cut
+
+sub prepare {
+  my $self  = shift;
+  my $sql   = shift;
+  my $cache = $self->{'_prepared'} ||= {};
+  my $sth   = $cache->{$sql}       ||= $self->dbh->prepare($sql) or die($self->errstr);
+  return $sth;
+}
+
 =head2 AutoCommit
 
-Pass through to  dbh->{'AutoCommit'} or dbh->{'AutoCommit'}=shift;
+Wrapper around dbh->{'AutoCommit'}
 
   $dbx->AutoCommit(1);
   &doSomething if $dbx->AutoCommit;
@@ -173,10 +196,10 @@ Pass through to  dbh->{'AutoCommit'} or dbh->{'AutoCommit'}=shift;
 For transactions that must complete together, I recommend
 
   { #block to keep local... well... local.
-    local $dbx->dbh->{"AutoCommit"}=0;
-    $dbx->insert($sql1, @bind1);
-    $dbx->update($sql2, @bind2);
-    $dbx->insert($sql3, @bind3);
+    local $dbx->dbh->{'AutoCommit'}=0;
+    $dbx->sqlinsert($sql1, @bind1);
+    $dbx->sqlupdate($sql2, @bind2);
+    $dbx->sqlinsert($sql3, @bind3);
   } #What is AutoCommit now?  Do you care?
 
 If AutoCommit reverts to true at the end of the block then DBI commits.  Else AutoCommit is still false and still not committed.  This allows higher layers to determine commit functionality.
@@ -193,14 +216,14 @@ sub AutoCommit {
 
 =head2 RaiseError
 
-Pass through to  dbh->{'RaiseError'} or dbh->{'RaiseError'}=shift;
+Wrapper around dbh->{'RaiseError'}
 
   $dbx->RaiseError(1);
   &doSomething if $dbx->RaiseError;
 
   { #local block
-    local $dbx->dbh->{"RaiseError"}=0;
-    $dbx->insert($sql, @bind); #do not die
+    local $dbx->dbh->{'RaiseError'}=0;
+    $dbx->sqlinsert($sql, @bind); #do not die
   }
 
 =cut
@@ -215,33 +238,15 @@ sub RaiseError {
 
 =head2 errstr
 
-Returns $DBI::errstr
+Wrapper around $DBI::errstr
 
-  $dbx->errstr;
+  my $err=$dbx->errstr;
 
 =cut
 
 sub errstr {$DBI::errstr};
 
-=head2 dbh
-
-Sets or returns the database handle object.
-
-  $dbx->dbh;
-  $dbx->dbh($dbh);  #if you already have a connection
-
-=cut
-
-sub dbh {
-  my $self=shift;
-  if (@_) {
-    $self->{'dbh'}=shift;
-    $self->{"_prepared"}=undef; #clear cache if we switch handles
-  }
-  return $self->{'dbh'};
-}
-
-=head1 METHODS (Read)
+=head1 METHODS (Read) - SQL
 
 =head2 sqlcursor
 
@@ -269,47 +274,50 @@ Scalar references are passed in and out with a hash bind.
 Direct Plug-in for L<SQL::Abstract> but no column alias support.
 
   my $sabs=SQL::Abstract->new;
-  my $sth=$dbx->sqlcursor($sabs->select($table, \@fields, \%where, \@sort));
+  my $sth=$dbx->sqlcursor($sabs->select($table, \@columns, \%where, \@sort));
 
 =cut
 
 sub sqlcursor {
   my $self=shift;
   my $sql=shift;
-  my $sth=$self->_prepared->{$sql};
-  unless ($sth) {
-    $sth=$self->dbh->prepare($sql)     or die($self->errstr);
-    #clear cache if over limit
-    $self->{"_prepared"}=undef if scalar(keys %{$self->_prepared}) > 16;
-    $self->_prepared->{$sql}=$sth;
-  }
+  my $sth=$self->prepare($sql);
   if (ref($_[0]) eq "ARRAY") {
-    $sth->execute(@{$_[0]})            or die($self->errstr);
+    my $bind_aref=shift;
+    $sth->execute(@$bind_aref) or die(&_error_string($self->errstr, $sql, sprintf("[%s]", join(", ", @$bind_aref)), "Array Reference"));
   } elsif (ref($_[0]) eq "HASH") {
-    foreach my $key (keys %{$_[0]}) {
-      next unless $sql=~m/:$key\b/;
-      if (ref($_[0]->{$key}) eq "SCALAR") {
-        $sth->bind_param_inout(":$key" => $_[0]->{$key}, 255);
+    my $bind_href=shift;
+    foreach my $key (keys %$bind_href) {
+      next unless $sql=~m/:$key\b/;                #TODO: comments are scanned so /* :foo */ is not supported here
+      if (ref($bind_href->{$key}) eq "SCALAR") {
+        $sth->bind_param_inout(":$key" => $bind_href->{$key}, 255);
       } else {
-        $sth->bind_param(":$key" => $_[0]->{$key});
+        $sth->bind_param(":$key" => $bind_href->{$key});
       }
-    } 
-    $sth->execute                      or die($self->errstr);
+    }
+    $sth->execute or die(&_error_string($self->errstr, $sql, sprintf("{%s}", join(", ", map {join("=>", $_ => $bind_href->{$_})} sort keys %$bind_href)), "Hash Reference"));
   } else {
-    $sth->execute(@_)                  or die($self->errstr);
+    my @bind=@_;
+    $sth->execute(@bind) or die(&_error_string($self->errstr, $sql, sprintf("(%s)", join(", ", @bind)), "List"));
   }
   return $sth;
-}
 
-sub _prepared {
-  my $self=shift;
-  $self->{"_prepared"}={} unless ref($self->{"_prepared"}) eq "HASH";
-  return $self->{"_prepared"};
+  sub _error_string {
+    my $err=shift;
+    my $sql=shift;
+    my $bind_str=shift;
+    my $type=shift;
+    if ($bind_str) {
+      return sprintf("Database Execute Error: %s\nSQL: %s\nBind(%s): %s\n", $err, $sql, $type, $bind_str);
+    } else {
+      return sprintf("Database Prepare Error: %s\nSQL: %s\n", $err, $sql);
+    }
+  }
 }
 
 =head2 sqlscalar
 
-Returns the SQL result as a scalar.
+Returns the first row first column value as a scalar.
 
 This works great for selecting one value.
 
@@ -437,6 +445,7 @@ sub _sqlarrayarray {
   my $name=$sth->{'NAME'}; #DBD::mysql must store this first
   my $row=[];
   my @rows=();
+  #TODO: replace with fetchall_arrayref
   while ($row=$sth->fetchrow_arrayref()) {
     push @rows, [@$row];
   }
@@ -515,6 +524,23 @@ sub _sqlarrayhash {
   return wantarray ? @rows : \@rows;
 }
 
+=head2 sqlarrayobject
+
+Returns the SQL result as an array of blessed hash objects in to the $class namespace.
+
+  my $array=$dbx->sqlarrayobject($class, $sql,  @parameters); #returns (bless({}, $class), ...)
+  my @array=$dbx->sqlarrayobject($class, $sql,  @parameters); #returns [bless({}, $class), ...]
+  my ($object)=$dbx->sqlarrayobject($class, $sql,  {id=>$id}); #$object is bless({}, $class)
+
+=cut
+
+sub sqlarrayobject {
+  my $self=shift;
+  my $class=shift or die("Error: The sqlarrayobject method requires a class parameter");
+  my @objects=map {bless($_, $class)} $self->sqlarrayhash(@_);
+  wantarray ? @objects : \@objects;
+}
+
 =head2 sqlsort (Oracle Specific?)
 
 Returns the SQL statement with the correct ORDER BY clause given a SQL statement (without an ORDER BY clause) and a signed integer on which column to sort.
@@ -525,7 +551,7 @@ Returns
 
   SELECT 1,'Z' FROM DUAL UNION SELECT 2,'A' FROM DUAL ORDER BY 2 DESC
 
-=cut 
+=cut
 
 sub sqlsort {
   my $self=shift;
@@ -534,7 +560,7 @@ sub sqlsort {
   if (defined($sort) and $sort=int($sort)) {
     my $column=abs($sort);
     my $direction = $sort < 0 ? "DESC" : "ASC";
-    return join " ", $sql, sprintf("ORDER BY %u %s", $column, $direction);  
+    return join " ", $sql, sprintf("ORDER BY %u %s", $column, $direction);
   } else {
     return $sql;
   }
@@ -557,42 +583,170 @@ sub sqlarrayarraynamesort {
   my $sql=shift;
   my $sort=shift;
   return $self->sqlarrayarrayname($self->sqlsort($sql, $sort), @_);
-} 
+}
 
-=head1 METHODS (Write)
+=head1 METHODS (Read) - SQL::Abstract
+
+=head2 abscursor
+
+Returns the prepared and executed SQL cursor.
+
+  my $sth=$dbx->abscursor($table, \@columns, \%where, \@order);
+
+=cut
+
+sub abscursor {
+  my $self=shift;
+  return $self->sqlcursor($self->abs->select(@_));
+}
+
+=head2 absscalar
+
+Returns the first row first column value as a scalar.
+
+  my $scalar=$dbx->absscalar($table, \@columns, \%where, \@order); #returns $
+
+=cut
+
+sub absscalar {
+  my $self=shift;
+  return $self->sqlscalar($self->abs->select(@_));
+}
+
+=head2 absarray
+
+Returns the SQL result as a array.
+
+This works great for selecting one value.
+
+  my @array=$dbx->absarray($table, \@columns, \%where, \@order); #returns ()
+  my $array=$dbx->absarray($table, \@columns, \%where, \@order); #returns []
+
+=cut
+
+sub absarray {
+  my $self=shift;
+  return $self->sqlarray($self->abs->select(@_));
+}
+
+=head2 abshash
+
+Returns the first two columns of the SQL result as a hash or hash reference {Key=>Value, Key=>Value, ...}
+
+  my $hash=$dbx->abshash($table, \@columns, \%where, \@order); #returns {}
+  my %hash=$dbx->abshash($table, \@columns, \%where, \@order); #returns ()
+
+=cut
+
+sub abshash {
+  my $self=shift;
+  return $self->sqlhash($self->abs->select(@_));
+}
+
+=head2 absarrayarray
+
+Returns the SQL result as an array or array ref of array references ([],[],...) or [[],[],...]
+
+  my $array=$dbx->absarrayarray($table, \@columns, \%where, \@order); #returns [[$,$,...],[],[],...]
+  my @array=$dbx->absarrayarray($table, \@columns, \%where, \@order); #returns ([$,$,...],[],[],...)
+
+=cut
+
+sub absarrayarray {
+  my $self=shift;
+  return $self->sqlarrayarray($self->abs->select(@_));
+}
+
+=head2 absarrayarrayname
+
+Returns the SQL result as an array or array ref of array references ([],[],...) or [[],[],...] where the first row contains an array reference to the column names
+
+  my $array=$dbx->absarrayarrayname($table, \@columns, \%where, \@order); #returns [[$,$,...],[],[],...]
+  my @array=$dbx->absarrayarrayname($table, \@columns, \%where, \@order); #returns ([$,$,...],[],[],...)
+
+=cut
+
+sub absarrayarrayname {
+  my $self=shift;
+  return $self->sqlarrayarrayname($self->abs->select(@_));
+}
+
+=head2 absarrayhash
+
+Returns the SQL result as an array or array ref of hash references ({},{},...) or [{},{},...]
+
+  my $array=$dbx->absarrayhash($table, \@columns, \%where, \@order); #returns [{},{},{},...]
+  my @array=$dbx->absarrayhash($table, \@columns, \%where, \@order); #returns ({},{},{},...)
+
+=cut
+
+sub absarrayhash {
+  my $self=shift;
+  return $self->sqlarrayhash($self->abs->select(@_));
+}
+
+=head2 absarrayhashname
+
+Returns the SQL result as an array or array ref of hash references ({},{},...) or [{},{},...]
+
+  my $array=$dbx->absarrayhashname($table, \@columns, \%where, \@order); #returns [[],{},{},...]
+  my @array=$dbx->absarrayhashname($table, \@columns, \%where, \@order); #returns ([],{},{},...)
+
+=cut
+
+sub absarrayhashname {
+  my $self=shift;
+  return $self->sqlarrayhashname($self->abs->select(@_));
+}
+
+=head2 absarrayobject
+
+Returns the SQL result as an array of blessed hash objects in to the $class namespace.
+
+  my $array=$dbx->absarrayobject($class, $table, \@columns, \%where, \@order); #returns (bless({}, $class), ...)
+  my @array=$dbx->absarrayobject($class, $table, \@columns, \%where, \@order); #returns [bless({}, $class), ...]
+
+=cut
+
+sub absarrayobject {
+  my $self=shift;
+  my $class=shift or die("Error: The absarrayobject method requires a class parameter");
+  my @objects=map {bless($_, $class)} $self->absarrayhash(@_);
+  wantarray ? @objects : \@objects;
+}
+
+=head1 METHODS (Write) - SQL
 
 Remember to commit or use AutoCommit
 
-Note: It appears that some drivers do not support the count of rows.  
+Note: It appears that some drivers do not support the count of rows.
 
-=head2 insert
+=head2 sqlinsert, insert
 
 Returns the number of rows inserted by the SQL statement.
 
-  my $rows=$dbx->insert( $sql,   @parameters);
-  my $rows=$dbx->insert( $sql,  \@parameters);
-  my $rows=$dbx->insert( $sql,  \%parameters);
-
-  my $sabs=SQL::Abstract->new;
-  my $rows=$dbx->insert($sabs->insert($table, \%field));
+  my $count=$dbx->sqlinsert( $sql,   @parameters);
+  my $count=$dbx->sqlinsert( $sql,  \@parameters);
+  my $count=$dbx->sqlinsert( $sql,  \%parameters);
 
 =cut
 
+*sqlinsert=\&update;
+
 *insert=\&update;
 
-=head2 update
+=head2 sqlupdate, update
 
 Returns the number of rows updated by the SQL statement.
 
-  my $rows=$dbx->update( $sql,   @parameters);
-  my $rows=$dbx->update( $sql,  \@parameters);
-  my $rows=$dbx->update( $sql,  \%parameters);
-
-  my $sabs=SQL::Abstract->new;
-  my $rows=$dbx->update($sabs->update($table, \%field, \%where));
+  my $count=$dbx->sqlupdate( $sql,   @parameters);
+  my $count=$dbx->sqlupdate( $sql,  \@parameters);
+  my $count=$dbx->sqlupdate( $sql,  \%parameters);
 
 =cut
 
+*sqlupdate=\&update;
+
 sub update {
   my $self=shift;
   my $sql=shift;
@@ -602,21 +756,20 @@ sub update {
   return $rows;
 }
 
-=head2 delete
+=head2 sqldelete, delete
 
 Returns the number of rows deleted by the SQL statement.
 
-  my $rows=$dbx->delete( $sql,   @parameters);
-  my $rows=$dbx->delete( $sql,  \@parameters);
-  my $rows=$dbx->delete( $sql,  \%parameters);
-
-  my $sabs=SQL::Abstract->new;
-  my $rows=$dbx->delete($sabs->delete($table, \%where));
+  my $count=$dbx->sqldelete($sql,   @parameters);
+  my $count=$dbx->sqldelete($sql,  \@parameters);
+  my $count=$dbx->sqldelete($sql,  \%parameters);
 
 Note: Some Oracle clients do not support row counts on delete instead the value appears to be a success code.
 
 =cut
 
+*sqldelete=\&update;
+
 *delete=\&update;
 
 =head2 execute, exec
@@ -624,28 +777,307 @@ Note: Some Oracle clients do not support row counts on delete instead the value
 Executes stored procedures.
 
   my $out;
-  my $rows=$dbx->execute($sql, $in, \$out);            #pass in/out vars as scalar reference
-  my $rows=$dbx->execute($sql, [$in, \$out]);
-  my $rows=$dbx->execute($sql, {in=>$in, out=>\$out});
+  my $return=$dbx->execute($sql, $in, \$out);            #pass in/out vars as scalar reference
+  my $return=$dbx->execute($sql, [$in, \$out]);
+  my $return=$dbx->execute($sql, {in=>$in, out=>\$out});
 
-Note: Currently update, insert, delete, and execute all point to the same method.  This may change in the future if we need to change the behavior of one method.  So, please use the correct method name for your function.
+Note: Currently sqlupdate, sqlinsert, sqldelete, and execute all point to the same method.  This may change in the future if we need to change the behavior of one method.  So, please use the correct method name for your function.
 
 =cut
 
 *execute=\&update;
 *exec=\&update;   #deprecated
 
-=head1 Get Info Methods
+=head1 METHODS (Write) - SQL::Abstract
+
+=head2 absinsert
+
+Returns the number of rows inserted.
+
+  my $count=$dbx->absinsert($table, \%column_values);
+
+=cut
+
+sub absinsert {
+  my $self=shift;
+  return $self->sqlinsert($self->abs->insert(@_));
+}
+
+=head2 absupdate
+
+Returns the number of rows updated.
+
+  my $count=$dbx->absupdate($table, \%column_values, \%where);
+
+=cut
+
+sub absupdate {
+  my $self=shift;
+  return $self->sqlupdate($self->abs->update(@_));
+}
+
+=head2 absdelete
+
+Returns the number of rows deleted.
+
+  my $count=$dbx->absdelete($table, \%where);
+
+=cut
+
+sub absdelete {
+  my $self=shift;
+  return $self->sqldelete($self->abs->delete(@_));
+}
+
+=head1 METHODS (Write) - Bulk - SQL
+
+=head2 bulksqlinsertarrayarray
+
+Insert records in bulk.
+
+  my @arrayarray=(
+                  [data1, $data2, $data3, $data4, ...],
+                  [@row_data_2],
+                  [@row_data_3], ...
+                 );
+  my $count=$dbx->bulksqlinsertarrayarray($sql, \@arrayarray);
+
+=cut
+
+sub bulksqlinsertarrayarray {
+  my $self         = shift;
+  my $sql          = shift or die('Error: sql required.');
+  my $arrayarray   = shift or die('Error: array of array references required.');
+  my $sth          = $self->prepare($sql);
+  my $size         = @$arrayarray;
+  my @tuple_status = ();
+  my $count        = $sth->execute_for_fetch( sub {shift @$arrayarray}, \@tuple_status);
+  unless ($count == $size) {
+    warn map {"$_\n"} @tuple_status; #TODO better error trapping...
+  }
+  return $count;
+}
+
+=head2 bulksqlinsertcursor
+
+Insert records in bulk.
+
+Step 1 select data from table 1 in database 1
+
+  my $sth1=$dbx1->sqlcursor('Select Col1 AS "ColA", Col2 AS "ColB", Col3 AS "ColC" from table1');
+
+Step 2 insert in to table 2 in database 2
+
+  my $count=$dbx2->bulksqlinsertcursor($sql, $sth1);
+
+Note: If you are inside a single database, it is much more efficient to use insert from select syntax. As there is no need for data to be transferred between the server and the client.
+
+=cut
+
+sub bulksqlinsertcursor {
+  my $self         = shift;
+  my $sql          = shift or die('Error: sql required.');
+  my $cursor       = shift or die('Error: cursor required.');
+  my $sth          = $self->prepare($sql);
+  my @tuple_status = ();
+  my $size         = 0;
+  my $count        = $sth->execute_for_fetch( sub {my $row=$cursor->fetchrow_arrayref; $size++ if $row; return $row}, \@tuple_status);
+  unless ($count == $size) {
+    warn Dumper \@tuple_status; #TODO better error trapping...
+  }
+  return $count;
+}
+
+=head2 bulksqlupdatearrayarray
+
+Update records in bulk.
+
+  my @arrayarray   = (
+                      [$data1, $data2, $data3, $data4, $id],
+                      [@row_data_2],
+                      [@row_data_3], ...
+                     );
+  my $count        = $dbx->bulksqlupdatearrayarray($sql, \@arrayarray);
+
+=cut
+
+sub bulksqlupdatearrayarray {
+  my $self         = shift;
+  my $sql          = shift or die('Error: sql required.');
+  my $arrayarray   = shift or die('Error: array of array references required.');
+  my $sth          = $self->prepare($sql);
+  my $size         = @$arrayarray;
+  my @tuple_status = ();
+  my $noerror      = $sth->execute_for_fetch( sub {shift @$arrayarray}, \@tuple_status);
+  warn("Warning: Atempted $size updates but only $noerror where successful.") unless $size == $noerror;
+  my $count        = sum grep {$_ > 0} @tuple_status;
+  return $count // 0;
+}
+
+=head1 METHODS (Write) - Bulk - SQL::Abstract-like
+
+These bulk methods do not use L<SQL::Abstract> but our own similar SQL insert and update methods.
+
+=head2 bulkabsinsertarrayarray
+
+Insert records in bulk.
+
+  my @columns=("Col1", "Col2", "Col3", "Col4", ...);
+  my @arrayarray=(
+                  [data1, $data2, $data3, $data4, ...],
+                  [@row_data_2],
+                  [@row_data_3], ...
+                 );
+  my $count=$dbx->bulkabsinsertarrayarray($table, \@columns, \@arrayarray);
+
+=cut
+
+sub bulkabsinsertarrayarray {
+  my $self         = shift;
+  my $table        = shift or die('Error: table name required.');
+  my $columns      = shift or die('Error: columns array reference required.');
+  my $arrayarray   = shift or die('Error: array of array references required.');
+  my $sql          = $self->_bulkinsert_sql($table => $columns);
+  return $self->bulksqlinsertarrayarray($sql, $arrayarray);
+}
+
+=head2 bulkabsinsertarrayhash
+
+Insert records in bulk.
+
+  my @columns=("Col1", "Col2", "Col3", "Col4", ...);                           #case sensative with respect to @arrayhash
+  my @arrayhash=(
+                 {C0l1=>data1, Col2=>$data2, Col3=>$data3, Col4=>$data4, ...}, #extra hash items ignored when sliced using @columns
+                 \%row_hash_data_2,
+                 \%row_hash_data_3, ...
+                );
+  my $count=$dbx->bulkabsinsertarrayhash($table, \@columns, \@arrayhash);
+
+=cut
+
+sub bulkabsinsertarrayhash {
+  my $self       = shift;
+  my $table      = shift or die("Error: table name required.");
+  my $columns    = shift or die("Error: columns array reference required.");
+  my $arrayhash  = shift or die("Error array of hash references required");
+  my @arrayarray = map {my %hash=%$_; my @slice=@hash{@$columns}; \@slice} @$arrayhash;
+  return $self->bulkabsinsertarrayarray($table, $columns, \@arrayarray);
+}
+
+=head2 bulkabsinsertcursor
+
+Insert records in bulk.
+
+Step 1 select data from table 1 in database 1
+
+  my $sth1=$dbx1->sqlcursor('Select Col1 AS "ColA", Col2 AS "ColB", Col3 AS "ColC" from table1');
+
+Step 2 insert in to table 2 in database 2
+
+  my $count=$dbx2->bulkabsinsertcursor($table2, $sth1);
+
+  my $count=$dbx2->bulkabsinsertcursor($table2, \@columns, $sth1); #if your DBD/API does not support column alias support
+
+Note: If you are inside a single database, it is much more efficient to use insert from select syntax. As no data needs to be transferred to and from the client.
+
+=cut
+
+sub bulkabsinsertcursor {
+  my $self         = shift;
+  my $table        = shift or die('Error: table name required.');
+  my $cursor       = pop   or die('Error: cursor required.');
+  my $columns      = shift || $cursor->{'NAME'};
+  my $sql          = $self->_bulkinsert_sql($table => $columns);
+  return $self->bulksqlinsertcursor($sql, $cursor);
+}
+
+#head2 _bulkinsert_sql
+#
+#Our own method since SQL::Abstract does not support ordered column values
+#
+#cut
+
+sub _bulkinsert_sql {
+  my $self=shift;
+  my $table=shift;
+  my $columns=shift;
+  my $sql=sprintf("INSERT INTO $table (%s) VALUES (%s)", join(',', @$columns), join(',', map {'?'} @$columns));
+  #warn "$sql\n";
+  return $sql;
+}
+
+=head2 bulkabsupdatearrayarray
+
+Update records in bulk.
+
+  my @setcolumns   = ("Col1", "Col2", "Col3", "Col4");
+  my @wherecolumns = ("ID");
+  my @arrayarray   = (
+                      [$data1, $data2, $data3, $data4, $id],
+                      [@row_data_2],
+                      [@row_data_3], ...
+                     );
+  my $count        = $dbx->bulkabsupdatearrayarray($table, \@setcolumns, \@wherecolumns, \@arrayarray);
+
+=cut
+
+sub bulkabsupdatearrayarray {
+  my $self         = shift;
+  my $table        = shift or die('Error: table name required.');
+  my $setcolumns   = shift or die('Error: set columns array reference required.');
+  my $wherecolumns = shift or die('Error: where columns array reference required.');
+  my $arrayarray   = shift;
+  my $sql          = $self->_bulkupdate_sql($table => $setcolumns, $wherecolumns);
+  return $self->bulksqlupdatearrayarray($sql, $arrayarray);
+}
+
+#head2 _bulkinsert_sql
+#
+#Our own method since SQL::Abstract does not support ordered column values
+#
+##cut
+
+sub _bulkupdate_sql {
+  my $self=shift;
+  my $table=shift;
+  my $setcolumns=shift;
+  my $wherecolumns=shift;
+  my $sql=sprintf("UPDATE $table SET %s WHERE %s", join(", ", map {"$_ = ?"} @$setcolumns), join(" AND ", map {"$_ = ?"} @$wherecolumns));
+  #warn "$sql\n";
+  return $sql;
+}
+
+=head1 Constructors
+
+=head2 abs
+
+Returns a L<SQL::Abstract> object
+
+=cut
+
+sub abs {
+  my $self=shift;
+  $self->{'abs'}=shift if @_;
+  unless (defined $self->{'abs'}) {
+    eval 'use SQL::Abstract'; #run time require so as not to require installation for all users
+    my $error=$@;
+    die($error) if $error;
+    $self->{'abs'}=SQL::Abstract->new;
+  }
+  return $self->{'abs'};
+}
+
+=head1 Methods (Informational)
 
 =head2 dbms_name
 
-Return the DBMS Name (e.g. Oracle)
+Return the DBMS Name (e.g. Oracle, MySQL, PostgreSQL)
 
 =cut
 
 sub dbms_name {shift->dbh->get_info(17)};
 
-=head1 Session Helpers
+=head1 Methods (Session Management)
 
 These methods allow the setting of Oracle session features that are available in the v$session table.  If other databases support these features, please let me know.  But, as it stands, these method are non operational unless SQL_DBMS_NAME is Oracle.
 
@@ -653,7 +1085,7 @@ These methods allow the setting of Oracle session features that are available in
 
 Sets and returns the v$session.module (Oracle) value.
 
-Note: Module is set for you by BDB::Oracle.  However you may set it however you'd like.  It should be set once after connection and left alone.
+Note: Module is set for you by DBD::Oracle.  However you may set it however you'd like.  It should be set once after connection and left alone.
 
   $dbx->module("perl@host");      #normally set by DBD::Oracle
   $dbx->module($module, $action); #can set initial action too.
@@ -690,7 +1122,7 @@ sub _set_module_sql {
 =head2 client_info
 
 Sets and returns the v$session.client_info (Oracle) value.
- 
+
   $dbx->client_info("Running From crontab");
   my $client_info=$dbx->client_info();
 
@@ -735,6 +1167,10 @@ Sets and returns the v$session.action (Oracle) value.
 
 Note: This should be updated fairly often. Every loop if it runs for more than 5 seconds and may end up in V$SQL_MONITOR.
 
+  while ($this) {
+    local $dbx->{'action'}="This Loop"; #tied to the database with a little Perl sugar
+  }
+
 =cut
 
 sub action {
@@ -769,9 +1205,9 @@ Sets and returns the v$session.client_identifier (Oracle) value.
   $dbx->client_identifier($login);
   my $client_identifier = $dbx->client_identifier();
 
-Note: This should be updated based on the login of the authenticated end user.  I use the client_info->{"realm"} if you have more than one authentication realm.
+Note: This should be updated based on the login of the authenticated end user.  I use the client_info->{'realm'} if you have more than one authentication realm.
 
-For auditing add this a to an update trigger
+For auditing add this to an update trigger
 
   new.UPDATED_USER = sys_context('USERENV', 'CLIENT_IDENTIFIER');
 
@@ -797,7 +1233,7 @@ sub _set_client_identifier_sql {
             --Package: $PACKAGE
             --Method: _client_identifier_sql
             BEGIN
-              DBMS_SESSION.SET_IDENTIFIER(client_id => ?); 
+              DBMS_SESSION.SET_IDENTIFIER(client_id => ?);
             END;
            };
 }
@@ -813,7 +1249,13 @@ sub _sys_context_userenv_sql {
 
 =head1 TODO
 
-Sort functions may not be portable.
+Sort functions sqlsort and sqlarrayarraynamesort may not be portable.
+
+Add some kind of capability to allow hash binds to bind as some native type rather than all strings.
+
+Hash binds scan comments for bind variables e.g. /* :variable */
+
+Improve error messages
 
 =head1 BUGS
 
@@ -843,7 +1285,7 @@ The full text of the license can be found in the LICENSE file included with this
 
 =head2 The Competition
 
-L<DBIx::DWIW>, L<DBIx::Wrapper>, L<DBIx::Simple>, L<Data::Table::fromSQL>, L<DBIx::Wrapper::VerySimple>
+L<DBIx::DWIW>, L<DBIx::Wrapper>, L<DBIx::Simple>, L<Data::Table::fromSQL>, L<DBIx::Wrapper::VerySimple>, L<DBIx::Raw>
 
 =head2 The Building Blocks
 
